from typing import List

from dokusan.entities import Cell, Sudoku


class Colors:
    BOLD = "\033[;1m"
    ORANGE = "\033[93m"
    ENDC = "\033[0m"


def plain(sudoku: List[List[int]]) -> str:
    result = "┌───┬───┬───╥───┬───┬───╥───┬───┬───┐\n"
    for i, row in enumerate(sudoku):
        for j, val in enumerate(row):
            if j == 0:
                result += "│ "
            if j == len(sudoku) - 1:
                border = " │"
            else:
                border = " ║ " if (j + 1) % 3 == 0 else " │ "
            result += f"{val}{border}" if val else f" {border}"
        result += "\n"
        if i == len(sudoku) - 1:
            result += "└───┴───┴───╨───┴───┴───╨───┴───┴───┘\n"
        elif (i + 1) % 3 == 0:
            result += "╞═══╪═══╪═══╬═══╪═══╪═══╬═══╪═══╪═══╡\n"
        else:
            result += "├───┼───┼───╫───┼───┼───╫───┼───┼───┤\n"
    return result


def colorful(sudoku: Sudoku) -> str:
    result = [_build_horizontal_border(sudoku, "┌", "─", "┬", "╥", "┐")]
    for i, row in enumerate(sudoku.rows()):
        result.append(_build_row(row))

        if i == len(sudoku.rows()) - 1:
            result.append(_build_horizontal_border(sudoku, "└", "─", "┴", "╨", "┘"))
        elif (i + 1) % 3 == 0:
            result.append(_build_horizontal_border(sudoku, "╞", "═", "╪", "╬", "╡"))
        else:
            result.append(_build_horizontal_border(sudoku, "├", "─", "┼", "╫", "┤"))

    return "".join(result)


def _build_horizontal_border(
    sudoku: Sudoku, right: str, middle: str, cell_sep: str, box_sep: str, left: str
) -> str:
    items = [f"{right}"]
    for i in range(sudoku.size):
        for j in range(sudoku.box_size.length * 3):
            items.append(f"{middle}")
        if (i + 1) == sudoku.size:
            items.append(f"{left}\n")
        elif (i + 1) % sudoku.box_size.width == 0:
            items.append(f"{box_sep}")
        else:
            items.append(f"{cell_sep}")
    return "".join(items)


def _build_row(row: List[Cell], size: int = 3) -> str:
    result = []
    for k in range(size):
        result.append("│")
        for j, cell in enumerate(row):
            start, end = k * 3, k * 3 + size
            for i in range(start, end):
                if cell.value and k == size // 2 and i == (start + end) // 2:
                    result.append(f" {Colors.BOLD}{cell.value}{Colors.ENDC} ")
                elif cell.candidates and k == i // 3 and (i + 1) in cell.candidates:
                    result.append(f" {Colors.ORANGE}{i + 1}{Colors.ENDC} ")
                else:
                    result.append("   ")
            result.append("║" if (j + 1) % size == 0 and j != len(row) - 1 else "│")
        result.append("\n")
    return "".join(result)



from dataclasses import dataclass, field
from typing import List, NamedTuple, Optional, Set, Tuple, Type, TypeVar

T = TypeVar("T", bound="Sudoku")


class Position(NamedTuple):
    row: int
    column: int
    box: int


class BoxSize(NamedTuple):
    width: int
    length: int

    def sequential(self, row: int, column: int) -> int:
        return (row // self.width) * self.width + (column // self.length)


@dataclass
class Cell:
    position: Position
    value: Optional[int] = None
    candidates: Set[int] = field(default_factory=set)

    def __post_init__(self):
        if self.value and self.candidates:
            raise ValueError("`value` and `candidates` attrs are mutually exclusive")


class Sudoku:
    def __init__(self, *cells: Cell, box_size: BoxSize = BoxSize(3, 3)):
        self.box_size = box_size
        self.size = box_size.width * box_size.length
        self._sudoku = {
            (i, j): Cell(position=Position(i, j, box_size.sequential(i, j)))
            for i in range(self.size)
            for j in range(self.size)
        }
        for cell in cells:
            self._sudoku[cell.position[:2]] = cell

    @classmethod
    def from_list(
        cls: Type[T], puzzle: List[List[int]], box_size: BoxSize = BoxSize(3, 3)
    ) -> T:
        cells = []
        for i, row in enumerate(puzzle):
            for j, value in enumerate(row):
                position = Position(i, j, box_size.sequential(i, j))
                cells.append(Cell(position=position, value=value if value else None,))
        return cls(*cells, box_size=box_size)

    def __getitem__(self, key: Tuple[int, int]) -> Cell:
        return self._sudoku[key]

    def update(self, cells: List[Cell]) -> None:
        self._sudoku.update((cell.position[:2], cell) for cell in cells)

    def cells(self) -> List[Cell]:
        return list(self._sudoku.values())

    def rows(self) -> List[List[Cell]]:
        return [[self[i, j] for j in range(self.size)] for i in range(self.size)]

    def columns(self) -> List[List[Cell]]:
        return [[self[j, i] for j in range(self.size)] for i in range(self.size)]

    def boxes(self) -> List[List[Cell]]:
        result: List[List[Cell]] = [[] for i in range(self.size)]
        for cell in self.cells():
            result[cell.position.box].append(cell)
        return result

    def groups(self) -> List[List[Cell]]:
        return self.rows() + self.columns() + self.boxes()

    def is_solved(self) -> bool:
        for group in self.groups():
            if len({cell.value for cell in group if cell.value}) != self.size:
                return False
        return True

    def is_valid(self) -> bool:
        for group in self.groups():
            values = [cell.value for cell in group if cell.value]
            if len(set(values)) != len(values):
                return False
        return True

    def intersection(self, *cells: Cell) -> List[Cell]:
        intersections = [
            {c.position for c in self.cells() if self.is_intersects(c, cell)}
            for cell in cells
        ]
        positions = set.intersection(*intersections)
        return [self[position.row, position.column] for position in positions]

    def is_intersects(self, cell_a: Cell, cell_b: Cell) -> bool:
        return cell_a.position != cell_b.position and (
            cell_a.position.row == cell_b.position.row
            or cell_a.position.column == cell_b.position.column
            or cell_a.position.box == cell_b.position.box
        )



# -*- coding: utf-8 -*-
from distutils.core import setup

package_dir = \
{'': 'src'}

packages = \
['dokusan']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'dokusan',
    'version': '0.1.0a1',
    'description': 'Sudoku solver with step-by-step guidance',
    'long_description': '========\nOverview\n========\n\nSudoku solver with step-by-step guidance\n\nThis lib is currently in pre-alpha phase,\nbut already can solve some simple and medium level sudoku.\n\nThe following code displays all steps leading to solution:\n\n.. code-block:: python\n\n    from dokusan import entities, techniques\n\n\n    class Unsolvable(Exception):\n        pass\n\n\n    def list_steps(sudoku: entities.Sudoku):\n        all_techniques = (\n            techniques.PencilMarking,\n            techniques.LoneSingle,\n            techniques.HiddenSingle,\n            techniques.NakedPair,\n            techniques.NakedTriplet,\n            techniques.LockedCandidate,\n            techniques.XYWing,\n            techniques.UniqueRectangle,\n        )\n        while not sudoku.is_solved():\n            for technique in all_techniques:\n                try:\n                    result = technique(sudoku).first()\n                except techniques.NotFound as exc:\n                    continue\n                else:\n                    sudoku.update(result.changes)\n                    yield result\n                    break\n                raise Unsolvable\n\n    _ = 0\n\n    sudoku = entities.Sudoku.from_list([\n        [_, _, _, _, 9, _, 1, _, _],\n        [_, _, _, _, _, 2, 3, _, _],\n        [_, _, 7, _, _, 1, 8, 2, 5],\n        [6, _, 4, _, 3, 8, 9, _, _],\n        [8, 1, _, _, _, _, _, _, _],\n        [_, _, 9, _, _, _, _, _, 8],\n        [1, 7, _, _, _, _, 6, _, _],\n        [9, _, _, _, 1, _, 7, 4, 3],\n        [4, _, 3, _, 6, _, _, _, 1],\n    ])\n\n    for step in list_steps(sudoku):\n        print(step.combination)\n',
    'author': 'Aleksei Maslakov',
    'author_email': 'lesha.maslakov@gmail.com',
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import itertools
import operator
from collections import Counter
from dataclasses import dataclass
from typing import Dict, Iterable, Iterator, List, Optional, Tuple

from dokusan.entities import Cell, Sudoku


class NotFound(Exception):
    pass


@dataclass
class Result:
    combination: Combination
    changes: List[Cell]


@dataclass
class Combination:
    name: str
    cells: List[Cell]
    values: List[int]

    def __str__(self) -> str:
        values = (str(value) for value in self.values)
        positions = (str((c.position.row, c.position.column)) for c in self.cells)
        return f"{self.name}: `{', '.join(values)}` at {', '.join(positions)}"


class Technique:
    def __init__(self, sudoku: Sudoku):
        self.sudoku = sudoku

    def __iter__(self) -> Iterator[Result]:
        return (
            Result(combination=combination, changes=changes)
            for combination in self._find()
            if (changes := self._get_changes(combination))
        )

    def first(self) -> Result:
        try:
            return next(iter(self))
        except StopIteration:
            raise NotFound("Not found")

    def _find(self) -> Iterator[Combination]:
        raise NotImplementedError

    def _get_changes(self, combination: Combination) -> List[Cell]:
        raise NotImplementedError


class PencilMarking(Technique):
    def _find(self) -> Iterator[Combination]:
        for cell in self.sudoku.cells():
            if not cell.value:
                yield Combination(
                    name="Pencil Marking", cells=[cell], values=[],
                )

    def _get_changes(self, combination: Combination) -> List[Cell]:
        result = []
        for cell in combination.cells:
            candidates = self._get_candidates(cell)
            if not cell.candidates or cell.candidates - candidates:
                result.append(Cell(position=cell.position, candidates=candidates))
        return result

    def _get_candidates(self, cell):
        intersection = self.sudoku.intersection(cell)
        all_values = set(range(1, self.sudoku.size + 1))
        return all_values - set(cell.value for cell in intersection if cell.value)


class LoneSingle(Technique):
    def _find(self) -> Iterator[Combination]:
        for cell in self.sudoku.cells():
            if len(cell.candidates) == 1:
                yield Combination(
                    name="Lone Single", cells=[cell], values=list(cell.candidates),
                )

    def _get_changes(self, combination: Combination) -> List[Cell]:
        eliminated = set(combination.values)
        single = Cell(
            position=combination.cells[0].position, value=combination.values[0],
        )
        return [
            Cell(position=cell.position, candidates=cell.candidates - eliminated)
            for cell in self.sudoku.intersection(single)
            if cell.candidates and cell.candidates & eliminated
        ] + [single]


class HiddenSingle(Technique):
    def _find(self) -> Iterator[Combination]:
        for group in self.sudoku.groups():
            candidate_map: Dict[int, List[Cell]] = {}
            for cell in group:
                if cell.candidates:
                    for candidate in cell.candidates:
                        candidate_map.setdefault(candidate, []).append(cell)

            for candidate, cells in candidate_map.items():
                if len(cells) == 1:
                    yield Combination(
                        name="Hidden Single", cells=cells, values=[candidate],
                    )

    def _get_changes(self, combination: Combination) -> List[Cell]:
        eliminated = set(combination.values)
        single = Cell(
            position=combination.cells[0].position, value=combination.values[0],
        )
        return [
            Cell(position=cell.position, candidates=cell.candidates - eliminated)
            for cell in self.sudoku.intersection(single)
            if cell.candidates and cell.candidates & eliminated
        ] + [single]


class NakedPair(Technique):
    def _find(self) -> Iterator[Combination]:
        for group in self.sudoku.groups():
            candidates_map: Dict[Tuple[int, ...], List[Cell]] = {}
            for cell in group:
                if cell.candidates:
                    candidates_map.setdefault(tuple(cell.candidates), []).append(cell)

            for candidates, cells in candidates_map.items():
                if len(candidates) == 2 and len(cells) == 2:
                    yield Combination(
                        name="Naked Pair", cells=cells, values=list(candidates),
                    )

    def _get_changes(self, combination: Combination) -> List[Cell]:
        eliminated = set(combination.values)
        return [
            Cell(position=cell.position, candidates=cell.candidates - eliminated)
            for cell in self.sudoku.intersection(*combination.cells)
            if cell.candidates and cell.candidates & eliminated
        ]


class NakedTriplet(Technique):
    def _find(self) -> Iterator[Combination]:
        for group in self.sudoku.groups():
            cells = [cell for cell in group if cell.candidates]
            counter = Counter(tuple(cell.candidates) for cell in cells)

            for pair, count in counter.items():
                if 2 <= count < 4:
                    triplet = [c for c in cells if len(c.candidates.union(pair)) < 4]
                    if len(triplet) == 3:
                        yield Combination(
                            name="Naked Triplet",
                            cells=triplet,
                            values=list(set.union(*[c.candidates for c in triplet])),
                        )

    def _get_changes(self, combination: Combination) -> List[Cell]:
        eliminated = set(combination.values)
        return [
            Cell(position=cell.position, candidates=cell.candidates - eliminated)
            for cell in self.sudoku.intersection(*combination.cells)
            if cell.candidates and cell.candidates & eliminated
        ]


class LockedCandidate(Technique):
    def _find(self) -> Iterator[Combination]:
        for group in self.sudoku.boxes():
            candidate_map: Dict[int, List[Cell]] = {}
            for cell in group:
                if cell.candidates:
                    for candidate in cell.candidates:
                        candidate_map.setdefault(candidate, []).append(cell)

            for candidate, cells in candidate_map.items():
                if len(cells) == 2:
                    yield Combination(
                        name="Locked Candidate", cells=cells, values=[candidate]
                    )

    def _get_changes(self, combination: Combination) -> List[Cell]:
        eliminated = set(combination.values)
        return [
            Cell(position=cell.position, candidates=cell.candidates - eliminated)
            for cell in self.sudoku.intersection(*combination.cells)
            if cell.candidates and cell.candidates & eliminated
        ]


class XYWing(Technique):
    def _find(self) -> Iterator[Combination]:
        cells = [cell for cell in self.sudoku.cells() if len(cell.candidates) == 2]
        for triplet in itertools.combinations(cells, 3):
            if self._is_xy_wing(triplet):
                wing = sorted(triplet, key=operator.attrgetter("position"))
                yield Combination(
                    name="XY Wing",
                    cells=wing,
                    values=list(set.intersection(*[c.candidates for c in wing[::2]])),
                )

    def _is_xy_wing(self, cells: Iterable[Cell]) -> bool:
        if len(set.union(*[c.candidates for c in cells])) != 3:
            return False
        combinations = tuple(itertools.combinations(cells, 2))
        if sum(self.sudoku.is_intersects(a, b) for a, b in combinations) != 2:
            return False
        if any(len(a.candidates & b.candidates) != 1 for a, b in combinations):
            return False
        return True

    def _get_changes(self, combination: Combination) -> List[Cell]:
        eliminated = set(combination.values)
        return [
            Cell(position=c.position, candidates=c.candidates - eliminated)
            for c in self.sudoku.intersection(*[x for x in combination.cells[::2]])
            if c.candidates and c.candidates & eliminated
        ]


class UniqueRectangle(Technique):
    def _find(self) -> Iterator[Combination]:
        cells = [cell for cell in self.sudoku.cells() if len(cell.candidates) == 2]
        for edges in itertools.combinations(cells, r=3):
            if self._is_edges(edges):
                rectangle = self._build_rectangle(edges)
                # waiting for https://github.com/python/mypy/issues/7316
                if rectangle is not None:
                    yield Combination(
                        name="Unique Rectangle",
                        cells=rectangle,
                        values=list(
                            set.intersection(*[c.candidates for c in rectangle])
                        ),
                    )

    def _is_edges(self, cells: Iterable[Cell]) -> bool:
        if len(set.intersection(*[c.candidates for c in cells])) != 2:
            return False
        combinations = tuple(itertools.combinations(cells, 2))
        if sum(a.position.row == b.position.row for a, b in combinations) != 1:
            return False
        if sum(a.position.column == b.position.column for a, b in combinations) != 1:
            return False
        if sum(a.position.box == b.position.box for a, b in combinations) != 1:
            return False
        return True

    def _build_rectangle(self, edges: Iterable[Cell]) -> Optional[List[Cell]]:
        rows = {edge.position.row for edge in edges}
        cols = {edge.position.column for edge in edges}
        cells = [self.sudoku[i, j] for i, j in itertools.product(rows, cols)]
        rectangle = [cell for cell in cells if cell.candidates]
        if len(rectangle) != 4:
            return None
        if len(set.intersection(*[m.candidates for m in rectangle])) != 2:
            return None
        return rectangle

    def _get_changes(self, combination: Combination) -> List[Cell]:
        return next(
            [Cell(position=edge_a.position, candidates=diff)]
            for edge_a, edge_b in itertools.combinations(combination.cells, 2)
            if (diff := edge_a.candidates - edge_b.candidates)
        )
